#include "Lib_Generic.h"
#include "glcdfont.h"
#include <avr/pgmspace.h>

void Adafruit_GFX()
{
  _width    = WIDTH;
  _height   = HEIGHT;
  rotation  = 0;
  cursor_y  = cursor_x    = 0;
  textsize  = 1;
  textcolor = textbgcolor = 0xFFFF;
  wrap      = true;
}

// Draw a circle outline
void drawCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color, int slave_id) {
  int16_t f = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x = 0;
  int16_t y = r;

  drawPixel(x0  , y0+r, color, slave_id);
  drawPixel(x0  , y0-r, color, slave_id);
  drawPixel(x0+r, y0  , color, slave_id);
  drawPixel(x0-r, y0  , color, slave_id);

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;
    }
    x++;
    ddF_x += 2;
    f += ddF_x;
  
    drawPixel(x0 + x, y0 + y, color, slave_id);
    drawPixel(x0 - x, y0 + y, color, slave_id);
    drawPixel(x0 + x, y0 - y, color, slave_id);
    drawPixel(x0 - x, y0 - y, color, slave_id);
    drawPixel(x0 + y, y0 + x, color, slave_id);
    drawPixel(x0 - y, y0 + x, color, slave_id);
    drawPixel(x0 + y, y0 - x, color, slave_id);
    drawPixel(x0 - y, y0 - x, color, slave_id);
  }
}

void drawCircleHelper( int16_t x0, int16_t y0,
               int16_t r, uint8_t cornername, uint16_t color, int slave_id) {
  int16_t f     = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x     = 0;
  int16_t y     = r;

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f     += ddF_y;
    }
    x++;
    ddF_x += 2;
    f     += ddF_x;
    if (cornername & 0x4) {
      drawPixel(x0 + x, y0 + y, color, slave_id);
      drawPixel(x0 + y, y0 + x, color, slave_id);
    } 
    if (cornername & 0x2) {
      drawPixel(x0 + x, y0 - y, color, slave_id);
      drawPixel(x0 + y, y0 - x, color, slave_id);
    }
    if (cornername & 0x8) {
      drawPixel(x0 - y, y0 + x, color, slave_id);
      drawPixel(x0 - x, y0 + y, color, slave_id);
    }
    if (cornername & 0x1) {
      drawPixel(x0 - y, y0 - x, color, slave_id);
      drawPixel(x0 - x, y0 - y, color, slave_id);
    }
  }
}

void fillCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color, int slave_id) {
  drawFastVLine(x0, y0-r, 2*r+1, color, slave_id);
  fillCircleHelper(x0, y0, r, 3, 0, color, slave_id);
}

// Used to do circles and roundrects
void fillCircleHelper(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, int16_t delta, uint16_t color, int slave_id) {

  int16_t f     = 1 - r;
  int16_t ddF_x = 1;
  int16_t ddF_y = -2 * r;
  int16_t x     = 0;
  int16_t y     = r;

  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f     += ddF_y;
    }
    x++;
    ddF_x += 2;
    f     += ddF_x;

    if (cornername & 0x1) {
      drawFastVLine(x0+x, y0-y, 2*y+1+delta, color, slave_id);
      drawFastVLine(x0+y, y0-x, 2*x+1+delta, color, slave_id);
    }
    if (cornername & 0x2) {
      drawFastVLine(x0-x, y0-y, 2*y+1+delta, color, slave_id);
      drawFastVLine(x0-y, y0-x, 2*x+1+delta, color, slave_id);
    }
  }
}

// Bresenham's algorithm - thx wikpedia
void drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t color, int slave_id) {
  int16_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    swap(x0, y0);
    swap(x1, y1);
  }

  if (x0 > x1) {
    swap(x0, x1);
    swap(y0, y1);
  }

  int16_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);

  int16_t err = dx / 2;
  int16_t ystep;

  if (y0 < y1) {
    ystep = 1;
  } else {
    ystep = -1;
  }

  for (; x0<=x1; x0++) {
    if (steep) {
      drawPixel(y0, x0, color, slave_id);
    } else {
      drawPixel(x0, y0, color, slave_id);
    }
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;
    }
  }
}

// Draw a rectangle
void drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color, int slave_id) {
  drawFastHLine(x, y, w, color, slave_id);
  drawFastHLine(x, y+h-1, w, color, slave_id);
  drawFastVLine(x, y, h, color, slave_id);
  drawFastVLine(x+w-1, y, h, color, slave_id);
}

// Draw a rounded rectangle
void drawRoundRect(int16_t x, int16_t y, int16_t w, int16_t h, int16_t r, uint16_t color, int slave_id) {
  // smarter version
  drawFastHLine(x+r  , y    , w-2*r, color, slave_id); // Top
  drawFastHLine(x+r  , y+h-1, w-2*r, color, slave_id); // Bottom
  drawFastVLine(x    , y+r  , h-2*r, color, slave_id); // Left
  drawFastVLine(x+w-1, y+r  , h-2*r, color, slave_id); // Right
  // draw four corners
  drawCircleHelper(x+r    , y+r    , r, 1, color, slave_id);
  drawCircleHelper(x+w-r-1, y+r    , r, 2, color, slave_id);
  drawCircleHelper(x+w-r-1, y+h-r-1, r, 4, color, slave_id);
  drawCircleHelper(x+r    , y+h-r-1, r, 8, color, slave_id);
}

// Fill a rounded rectangle
void fillRoundRect(int16_t x, int16_t y, int16_t w, int16_t h, int16_t r, uint16_t color, int slave_id) {
  // smarter version
  fillRect(x+r, y, w-2*r, h, color, slave_id);

  // draw four corners
  fillCircleHelper(x+w-r-1, y+r, r, 1, h-2*r-1, color, slave_id);
  fillCircleHelper(x+r    , y+r, r, 2, h-2*r-1, color, slave_id);
}

// Draw a triangle
void drawTriangle(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color, int slave_id) {
  drawLine(x0, y0, x1, y1, color, slave_id);
  drawLine(x1, y1, x2, y2, color, slave_id);
  drawLine(x2, y2, x0, y0, color, slave_id);
}

// Fill a triangle
void fillTriangle ( int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color, int slave_id) {

  int16_t a, b, y, last;

  // Sort coordinates by Y order (y2 >= y1 >= y0)
  if (y0 > y1) {
    swap(y0, y1); swap(x0, x1);
  }
  if (y1 > y2) {
    swap(y2, y1); swap(x2, x1);
  }
  if (y0 > y1) {
    swap(y0, y1); swap(x0, x1);
  }

  if(y0 == y2) { // Handle awkward all-on-same-line case as its own thing
    a = b = x0;
    if(x1 < a)      a = x1;
    else if(x1 > b) b = x1;
    if(x2 < a)      a = x2;
    else if(x2 > b) b = x2;
    drawFastHLine(a, y0, b-a+1, color, slave_id);
    return;
  }

  int16_t
    dx01 = x1 - x0,
    dy01 = y1 - y0,
    dx02 = x2 - x0,
    dy02 = y2 - y0,
    dx12 = x2 - x1,
    dy12 = y2 - y1;
  int32_t
    sa   = 0,
    sb   = 0;

  // For upper part of triangle, find scanline crossings for segments
  // 0-1 and 0-2.  If y1=y2 (flat-bottomed triangle), the scanline y1
  // is included here (and second loop will be skipped, avoiding a /0
  // error there), otherwise scanline y1 is skipped here and handled
  // in the second loop...which also avoids a /0 error here if y0=y1
  // (flat-topped triangle).
  if(y1 == y2) last = y1;   // Include y1 scanline
  else         last = y1-1; // Skip it

  for(y=y0; y<=last; y++) {
    a   = x0 + sa / dy01;
    b   = x0 + sb / dy02;
    sa += dx01;
    sb += dx02;
    /* longhand:
    a = x0 + (x1 - x0) * (y - y0) / (y1 - y0);
    b = x0 + (x2 - x0) * (y - y0) / (y2 - y0);
    */
    if(a > b) swap(a,b);
    drawFastHLine(a, y, b-a+1, color, slave_id);
  }

  // For lower part of triangle, find scanline crossings for segments
  // 0-2 and 1-2.  This loop is skipped if y1=y2.
  sa = dx12 * (y - y1);
  sb = dx02 * (y - y0);
  for(; y<=y2; y++) {
    a   = x1 + sa / dy12;
    b   = x0 + sb / dy02;
    sa += dx12;
    sb += dx02;
    /* longhand:
    a = x1 + (x2 - x1) * (y - y1) / (y2 - y1);
    b = x0 + (x2 - x0) * (y - y0) / (y2 - y0);
    */
    if(a > b) swap(a,b);
    drawFastHLine(a, y, b-a+1, color, slave_id);
  }
}

void drawBitmap(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, int slave_id) {

  int16_t i, j, byteWidth = (w + 7) / 8;

  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(pgm_read_byte(bitmap + j * byteWidth + i / 8) & (128 >> (i & 7))) {
        drawPixel(x+i, y+j, color, slave_id);
      }
    }
  }
}

// Draw a 1-bit color bitmap at the specified x, y position from the
// provided bitmap buffer (must be PROGMEM memory) using color as the
// foreground color and bg as the background color.
void drawBitmapBg(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg, int slave_id) {

  int16_t i, j, byteWidth = (w + 7) / 8;
  
  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(pgm_read_byte(bitmap + j * byteWidth + i / 8) & (128 >> (i & 7))) {
        drawPixel(x+i, y+j, color, slave_id);
      }
      else {
      	drawPixel(x+i, y+j, bg, slave_id);
      }
    }
  }
}

//Draw XBitMap Files (*.xbm), exported from GIMP,
//Usage: Export from GIMP to *.xbm, rename *.xbm to *.c and open in editor.
//C Array can be directly used with this function
void drawXBitmap(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, int slave_id) {
  
  int16_t i, j, byteWidth = (w + 7) / 8;
  
  for(j=0; j<h; j++) {
    for(i=0; i<w; i++ ) {
      if(pgm_read_byte(bitmap + j * byteWidth + i / 8) & (1 << (i % 8))) {
        drawPixel(x+i, y+j, color, slave_id);
      }
    }
  }
}

void write(uint8_t c, int slave_id) {
  if (c == '\n') {
    cursor_y += textsize*8;
    cursor_x  = 0;
  } else if (c == '\r') {
    // skip em
  } else {
    drawChar(cursor_x, cursor_y, c, textcolor, textbgcolor, textsize, slave_id);
    cursor_x += textsize*6;
    if (wrap && (cursor_x > (_width - textsize*6))) {
      cursor_y += textsize*8;
      cursor_x = 0;
    }
  }
}

// Draw a character
void drawChar(int16_t x, int16_t y, unsigned char c, uint16_t color, uint16_t bg, uint8_t size, int slave_id) {

  if((x >= _width)            || // Clip right
     (y >= _height)           || // Clip bottom
     ((x + 6 * size - 1) < 0) || // Clip left
     ((y + 8 * size - 1) < 0))   // Clip top
    return;

  for (int8_t i=0; i<6; i++ ) {
    uint8_t line;
    if (i == 5) 
      line = 0x0;
    else 
      line = pgm_read_byte(font+(c*5)+i);
    for (int8_t j = 0; j<8; j++) {
      if (line & 0x1) {
        if (size == 1) // default size
          drawPixel(x+i, y+j, color, slave_id);
        else {  // big size
          fillRect(x+(i*size), y+(j*size), size, size, color, slave_id);
        } 
      } else if (bg != color) {
        if (size == 1) // default size
          drawPixel(x+i, y+j, bg, slave_id);
        else {  // big size
          fillRect(x+i*size, y+j*size, size, size, bg, slave_id);
        }
      }
      line >>= 1;
    }
  }
}

void setCursor(int16_t x, int16_t y, int slave_id) {
  cursor_x = x;
  cursor_y = y;
}

void setTextSize(uint8_t s, int slave_id) {
  textsize = (s > 0) ? s : 1;
}

void setTextColor(uint16_t c, int slave_id) {
  // For 'transparent' background, we'll set the bg 
  // to the same as fg instead of using a flag
  textcolor = textbgcolor = c;
}

void setTextColorBg(uint16_t c, uint16_t b, int slave_id) {
  textcolor   = c;
  textbgcolor = b; 
}

void setTextWrap(boolean w, int slave_id) {
  wrap = w;
}

uint8_t getRotation(void) {
  return rotation;
}


// Return the size of the display (per current rotation)
int16_t width(void) {
  return _width;
}
 
int16_t height(void) {
  return _height;
}


inline uint16_t swapcolor(uint16_t x) { 
  return (x << 11) | (x & 0x07E0) | (x >> 11);
}
void delay_ms(uint16_t count) {
	while(count--) {
		_delay_ms(1);

	}
}

void CS_SELECT(int slave_id)
{
	if (slave_id == SLAVE1)
	{
		if (!(SPI_PORT & (1 << SPI_CS2)))//other one is selected (0)
		SPI_PORT |= (1 << SPI_CS2);//make it high
		SPI_PORT &= ~(1<<SPI_CS1);//make this low
	}
	else
	{
		if (!(SPI_PORT & (1 << SPI_CS1)))//other one is selected (0)
		SPI_PORT |= (1 << SPI_CS1);//make it high
		SPI_PORT &= ~(1<<SPI_CS2);//make this low
	}
}
void CS_DESELECT(int slave_id)
{
	if (slave_id == SLAVE1)
	{
		SPI_PORT |= (1<<SPI_CS1);//make this high
	}
	else
	{
		SPI_PORT |= (1<<SPI_CS2);//make this high
	}
}

void SPI_init()
{
	PORTB = 0;
	// Set spi direction
	SPI_DDR |= (1 << SPI_MOSI) | (1 << SPI_SCK) | (1 << SPI_CS1) | (1 << SPI_CS2) | (1 << SPI_DC) | (1 << SPI_RST);
	SPI_DDR |= (1 << SPI_DC);
	
	// Initialize SPI port, SPI enabled as master
	SPCR0 = (1 << SPE0) | (1 << MSTR0);
	// spi speed /2
	SPSR0 |= (1 << SPI2X0);
	//MSB FIRST
	SPCR0 &= ~(1 << DORD0);
	
	//MOSI 0
	PORTB &= ~(1<<SPI_MOSI);
	//CLK 0	
	PORTB &= ~(1<<SPI_SCK);
	
	int i;
	for (i=1; i<=1; i++)
	{
		CS_SELECT(i);
	
		//TOGGLE RESET
		PORTB |= (1<<SPI_RST);
		delay_ms(100);	
		PORTB &= ~(1<<SPI_RST);
		delay_ms(100);
		PORTB |= (1<<SPI_RST);
		delay_ms(100);
	}	
}

inline void spiwrite(uint8_t c) {

  SPDR0 = c;
  while(!(SPSR0 & (1 << SPIF0)));
  (void)SPDR0;
}

uint8_t SPI_receive(void)
{
	return 0;
	SPDR0 = 0xFF;
	while(!(SPSR0 & (1 << SPIF0)));
	return SPDR0;
}

void writecommand(uint8_t c, int slave_id) {
  RS_SELECT();
  CS_SELECT(slave_id);

  spiwrite(c);
  CS_DESELECT(slave_id);
  SPI_receive();
}

void writedata(uint8_t c, int slave_id) {
  RS_DESELECT();
  CS_SELECT(slave_id);
  spiwrite(c);
  
  CS_DESELECT(slave_id);
  SPI_receive();
} 

// Rather than a bazillion writecommand() and writedata() calls, screen
// initialization commands and arguments are organized in these tables
// stored in PROGMEM.  The table may look bulky, but that's mostly the
// formatting -- storage-wise this is hundreds of bytes more compact
// than the equivalent code.  Companion function follows.
#define DELAY 0x80
static const uint8_t PROGMEM
  Rcmd1[] = {                 // Init for 7735R, part 1 (red or green tab)
    15,                       // 15 commands in list:
    SWRESET,   DELAY,  //  1: Software reset, 0 args, w/delay
      10,                    //     150 ms delay
    SLPOUT ,   DELAY,  //  2: Out of sleep mode, 0 args, w/delay
      10,                    //     500 ms delay
    FRMCTR1, 3      ,  //  3: Frame rate ctrl - normal mode, 3 args:
      0x01, 0x2C, 0x2D,       //     Rate = fosc/(1x2+40) * (LINE+2C+2D)
    FRMCTR2, 3      ,  //  4: Frame rate control - idle mode, 3 args:
      0x01, 0x2C, 0x2D,       //     Rate = fosc/(1x2+40) * (LINE+2C+2D)
    FRMCTR3, 6      ,  //  5: Frame rate ctrl - partial mode, 6 args:
      0x01, 0x2C, 0x2D,       //     Dot inversion mode
      0x01, 0x2C, 0x2D,       //     Line inversion mode
    INVCTR , 1      ,  //  6: Display inversion ctrl, 1 arg, no delay:
      0x07,                   //     No inversion
    PWCTR1 , 3      ,  //  7: Power control, 3 args, no delay:
      0xA2,
      0x02,                   //     -4.6V
      0x84,                   //     AUTO mode
    PWCTR2 , 1      ,  //  8: Power control, 1 arg, no delay:
      0xC5,                   //     VGH25 = 2.4C VGSEL = -10 VGH = 3 * AVDD
    PWCTR3 , 2      ,  //  9: Power control, 2 args, no delay:
      0x0A,                   //     Opamp current small
      0x00,                   //     Boost frequency
    PWCTR4 , 2      ,  // 10: Power control, 2 args, no delay:
      0x8A,                   //     BCLK/2, Opamp current small & Medium low
      0x2A,  
    PWCTR5 , 2      ,  // 11: Power control, 2 args, no delay:
      0x8A, 0xEE,
    VMCTR1 , 1      ,  // 12: Power control, 1 arg, no delay:
      0x0E,
    INVOFF , 0      ,  // 13: Don't invert display, no args, no delay
    MADCTL , 1      ,  // 14: Memory access control (directions), 1 arg:
      0xC8,                   //     row addr/col addr, bottom to top refresh
    COLMOD , 1      ,  // 15: set color mode, 1 arg, no delay:
      0x05 },                 //     16-bit color
  Rcmd2green[] = {            // Init for 7735R, part 2 (green tab only)
    2,                        //  2 commands in list:
    CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x02,             //     XSTART = 0
      0x00, 0x7F+0x02,        //     XEND = 127
    RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x01,             //     XSTART = 0
      0x00, 0x9F+0x01 },      //     XEND = 159
  Rcmd2red[] = {              // Init for 7735R, part 2 (red tab only)
    2,                        //  2 commands in list:
    CASET  , 4      ,  //  1: Column addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x7F,             //     XEND = 127
    RASET  , 4      ,  //  2: Row addr set, 4 args, no delay:
      0x00, 0x00,             //     XSTART = 0
      0x00, 0x9F },           //     XEND = 159
  Rcmd3[] = {                 // Init for 7735R, part 3 (red or green tab)
    4,                        //  4 commands in list:
    GMCTRP1, 16      , //  1: Magical unicorn dust, 16 args, no delay:
      0x02, 0x1c, 0x07, 0x12,
      0x37, 0x32, 0x29, 0x2d,
      0x29, 0x25, 0x2B, 0x39,
      0x00, 0x01, 0x03, 0x10,
    GMCTRN1, 16      , //  2: Sparkles and rainbows, 16 args, no delay:
      0x03, 0x1d, 0x07, 0x06,
      0x2E, 0x2C, 0x29, 0x2D,
      0x2E, 0x2E, 0x37, 0x3F,
      0x00, 0x00, 0x02, 0x10,
    NORON  ,    DELAY, //  3: Normal display on, no args, w/delay
      10,                     //     10 ms delay
    DISPON ,    DELAY, //  4: Main screen turn on, no args w/delay
      10 };                  //     100 ms delay

// Companion code to the above tables.  Reads and issues
// a series of LCD commands stored in PROGMEM byte array.
void commandList(const uint8_t *addr, int slave_id) {

  uint8_t  numCommands, numArgs;
  uint16_t ms;

  numCommands = pgm_read_byte(addr++);   // Number of commands to follow
  while(numCommands--) {                 // For each command...
    writecommand(pgm_read_byte(addr++), slave_id); //   Read, issue command
    numArgs  = pgm_read_byte(addr++);    //   Number of args to follow
    ms       = numArgs & DELAY;          //   If hibit set, delay follows args
    numArgs &= ~DELAY;                   //   Mask out delay bit
    while(numArgs--) {                   //   For each argument...
      writedata(pgm_read_byte(addr++), slave_id);  //     Read, issue argument
    }

    if(ms) {
      ms = pgm_read_byte(addr++); // Read post-command delay time (ms)
      if(ms == 255) ms = 500;     // If 255, delay for 500 ms
      delay_ms(ms);
    }
  }
}

// Initialization code common to both 'B' and 'R' type displays
void commonInit(const uint8_t *cmdList) 
{
  Adafruit_GFX();	
  colstart  = rowstart = 0; // May be overridden in init func

  SPI_init();
  int i;
  for (i=1; i<=2; i++)
  {
	  if(cmdList)
		commandList(cmdList, i);	
  }
}

void initR() {
  commonInit(Rcmd1);
  int i;
  for (i=1; i<=2; i++)
  {
	  commandList(Rcmd2red, i);
	  commandList(Rcmd3, i);
	  writecommand(MADCTL, i);
	  writedata(0xC0, i);
  }
}

void setAddrWindow(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, int slave_id) {

  writecommand(CASET, slave_id); // Column addr set
  writedata(0x00, slave_id);
  writedata(x0+colstart, slave_id);     // XSTART 
  writedata(0x00, slave_id);
  writedata(x1+colstart, slave_id);     // XEND

  writecommand(RASET, slave_id); // Row addr set
  writedata(0x00, slave_id);
  writedata(y0+rowstart, slave_id);     // YSTART
  writedata(0x00, slave_id);
  writedata(y1+rowstart, slave_id);     // YEND

  writecommand(RAMWR, slave_id); // write to RAM
}

void pushColor(uint16_t color, int slave_id) {
  RS_DESELECT();
  CS_SELECT(slave_id);
  
  spiwrite(color >> 8);
  spiwrite(color);

  //*csport |= cspinmask;
  CS_DESELECT(slave_id);
  SPI_receive();
}

void drawPixel(int16_t x, int16_t y, uint16_t color, int slave_id) {

  if((x < 0) ||(x >= _width) || (y < 0) || (y >= _height)) return;

  setAddrWindow(x,y,x+1,y+1, slave_id);

  RS_DESELECT();
  CS_SELECT(slave_id);
  
  spiwrite(color >> 8);
  spiwrite(color);
  
  CS_DESELECT(slave_id);
  SPI_receive();
}

void drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color, int slave_id) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;
  if((y+h-1) >= _height) h = _height-y;
  setAddrWindow(x, y, x, y+h-1, slave_id);

  uint8_t hi = color >> 8, lo = color;
  RS_DESELECT();
  CS_SELECT(slave_id);
  while (h--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  CS_DESELECT(slave_id);
  SPI_receive();
}

void drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color, int slave_id) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;
  if((x+w-1) >= _width)  w = _width-x;
  setAddrWindow(x, y, x+w-1, y, slave_id);

  uint8_t hi = color >> 8, lo = color;
  RS_DESELECT();
 CS_SELECT(slave_id);
  while (w--) {
    spiwrite(hi);
    spiwrite(lo);
  }
  CS_DESELECT(slave_id);
  SPI_receive();
}

void fillScreen(uint16_t color, int slave_id) {
  fillRect(0, 0,  _width, _height, color, slave_id);
}

// fill a rectangle
void fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color, int slave_id) {

  // rudimentary clipping (drawChar w/big text requires this)
  if((x >= _width) || (y >= _height)) return;
  if((x + w - 1) >= _width)  w = _width  - x;
  if((y + h - 1) >= _height) h = _height - y;

  setAddrWindow(x, y, x+w-1, y+h-1, slave_id);

  uint8_t hi = color >> 8, lo = color;
  RS_DESELECT();
  CS_SELECT(slave_id);
  for(y=h; y>0; y--) {
    for(x=w; x>0; x--) {
      spiwrite(hi);
      spiwrite(lo);
    }
  }

  CS_DESELECT(slave_id);
  SPI_receive();
}

// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t Color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

#define MADCTL_MY  0x80
#define MADCTL_MX  0x40
#define MADCTL_MV  0x20
#define MADCTL_ML  0x10
#define MADCTL_RGB 0x00
#define MADCTL_BGR 0x08
#define MADCTL_MH  0x04

void setRotation(uint8_t m, int slave_id) {

  writecommand(MADCTL, slave_id);
  rotation = m % 4; // can't be higher than 3
  switch (rotation) {
   case 0:
       writedata(MADCTL_MX | MADCTL_MY | MADCTL_RGB, slave_id);
     _width  = TFTWIDTH;
     _height = TFTHEIGHT_18;

     break;
   case 1:
       writedata(MADCTL_MY | MADCTL_MV | MADCTL_RGB, slave_id);
	 
     _width = TFTHEIGHT_18;
     _height = TFTWIDTH;
     break;
  case 2:
       writedata(MADCTL_RGB, slave_id);
     _width  = TFTWIDTH;
     _height = TFTHEIGHT_18;

    break;
   case 3:
       writedata(MADCTL_MX | MADCTL_MV | MADCTL_RGB, slave_id);
     _width = TFTHEIGHT_18;
     _height = TFTWIDTH;
     break;
  }
  SPI_receive();
}

void invertDisplay(boolean i, int slave_id) {
  writecommand(i ? INVON : INVOFF, slave_id);
}
void PrintString(char *str, int x, int y, uint16_t color, uint16_t bg, uint8_t size, int slave_id)
{
	int x_aux = x;
	while(*str)
	{
		drawChar(x_aux, y, *str, color, bg, size, slave_id);
		x_aux += size*6;
		str++;
	}
}
