#define F_CPU 16000000L
#include <util/delay.h>
#include <limits.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <avr/eeprom.h>
#include "Lib_Generic.h"

#define SLAVEID	1
#define CELLS_PER_ROW 8
#define CELL_SIZE 16
#define INITIAL_BOMBS_COUNT 2
#define LEVEL_INCREASE 3

#define UNOPENED_CELL_BACKGROUND  0x333333UL
#define UNOPENED_CELL_MARGIN 0x293546UL
#define OPENED_CELL_BACKGROUND 0x0UL
#define SELECTED_CELL_BACKGROUND 0x006d5bUL
#define BOMB_ACTIVATED_BACKGROUND 0x0UL
#define BOMB_FLAGGED_BACKGROUND 0x0UL
#define FLAGGED_CELL_BACKGROUND 0x0UL
#define FLAGGED_CELL_MARGIN 0x293546UL
#define FLAGGED_CELL_FLAG RED
#define BOMBS_COUNT_TEXT BLUE

#define BTN1 PC2
#define BTN2 PC4
#define BTN3 PC1
#define BTN4 PC3
#define BTN5 PC0
#define BTN6 PC5

typedef enum
{
	CELL_OPENED = 1<<2,
	CELL_SELECTED = 1<<3,
	CELL_BOMB_ACTIVATED = 1<<4,
	CELL_BOMB_FLAGGED = 1<<5,
	CELL_FLAGGED = 1<<6
}CELL_STATE;

int cell_bombs[CELLS_PER_ROW][CELLS_PER_ROW] = {0};//0-n number of bombs in vecinity, -1 for bomb there
CELL_STATE cell_states[CELLS_PER_ROW][CELLS_PER_ROW] = {0};

int left_bombs = INITIAL_BOMBS_COUNT;
int level = 1;
volatile int cursor_xpos=0;
volatile int cursor_ypos=0;
int last_cursor_xpos=0;
int last_cursor_ypos=0;

int selected_cell=0;
int flagged_cell=0;
volatile int lost_game=0;
volatile int BTN1_LAST=0, BTN2_LAST=0, BTN3_LAST=0, BTN4_LAST=0, BTN5_LAST=0, BTN6_LAST=0;

void DrawCell(int x, int y)
{
	if (cell_states[x][y] & CELL_SELECTED && !(cell_states[x][y] & CELL_BOMB_ACTIVATED))
	{
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, SELECTED_CELL_BACKGROUND, SLAVEID);
		return;
	}

	if (cell_states[x][y] & CELL_BOMB_ACTIVATED)
	{
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, BOMB_ACTIVATED_BACKGROUND, SLAVEID);
		fillCircle(x*CELL_SIZE+CELL_SIZE/2, y*CELL_SIZE+CELL_SIZE/2, CELL_SIZE/2-2, RED, SLAVEID);
	}
	else if (cell_states[x][y] & CELL_BOMB_FLAGGED)
	{
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, BOMB_FLAGGED_BACKGROUND, SLAVEID);
		fillCircle(x*CELL_SIZE+CELL_SIZE/2, y*CELL_SIZE+CELL_SIZE/2, CELL_SIZE/2-2, GREEN, SLAVEID);
	}
	else if (cell_states[x][y] & CELL_FLAGGED)
	{
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, FLAGGED_CELL_BACKGROUND, SLAVEID);
		drawRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, FLAGGED_CELL_MARGIN, SLAVEID);		
		fillRect(x*CELL_SIZE+4, y*CELL_SIZE+4, 2, 9, FLAGGED_CELL_FLAG, SLAVEID);
		fillRect(x*CELL_SIZE+4, y*CELL_SIZE+4, 5, 2, FLAGGED_CELL_FLAG, SLAVEID);
	}
	else if (cell_states[x][y] & CELL_OPENED)
	{
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, OPENED_CELL_BACKGROUND, SLAVEID);
		//has nr of bombs
		if (cell_bombs[x][y] > 0)
		{
			drawChar(x*CELL_SIZE+3, y*CELL_SIZE, '0'+cell_bombs[x][y], BOMBS_COUNT_TEXT, OPENED_CELL_BACKGROUND, 2, SLAVEID);
		}
	}
	else
	{
	    //UNOPENED
		fillRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, UNOPENED_CELL_BACKGROUND, SLAVEID);
		drawRect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE, UNOPENED_CELL_MARGIN, SLAVEID);
	}
}
void UpdateScore()
{
	const char* text1 = "Bombs left: ";
	const char* text2 = "Level:      ";

	char bombs_left_str[3];
	char level_str[3];
	int cnt = 0;
	int offsetx = 5;

	fillRect(0, CELLS_PER_ROW*CELL_SIZE, 128, 160-CELLS_PER_ROW*CELL_SIZE, BLACK, SLAVEID);

	while(*(text1+cnt))
	{
		drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+2, *(text1+cnt), RED, BLACK, 1, SLAVEID);
		offsetx += 8;
		cnt++;
	}
	sprintf(bombs_left_str, "%.2d", left_bombs);
	drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+2, bombs_left_str[0], RED, BLACK, 1, SLAVEID);
	offsetx += 8;
	drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+2, bombs_left_str[1], RED, BLACK, 1, SLAVEID);

	offsetx = 5;
	cnt = 0;
	while(*(text2+cnt))
	{
		drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+12, *(text2+cnt), RED, BLACK, 1, SLAVEID);
		offsetx += 8;
		cnt++;
	}
	sprintf(level_str, "%.2d", level);
	drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+12, level_str[0], RED, BLACK, 1, SLAVEID);
	offsetx += 8;
	drawChar(offsetx, CELLS_PER_ROW*CELL_SIZE+12, level_str[1], RED, BLACK, 1, SLAVEID);
	
}
void DrawTable()
{
	for (int i=0; i<CELLS_PER_ROW; i++)
	{
		for (int j=0; j<CELLS_PER_ROW; j++)
		{			
			DrawCell(i, j);
		}
	}
}
void GenerateTable()
{
	int sel_x;
	int sel_y;

	for (int i=0; i<left_bombs; i++)
	{
		do
		{
			sel_x = rand()%CELLS_PER_ROW;
			sel_y = rand()%CELLS_PER_ROW;
		}while (cell_bombs[sel_x][sel_y] == -1);
		cell_bombs[sel_x][sel_y] = -1;
	}

	for (int x=0; x<CELLS_PER_ROW; x++)
	{
		for (int y=0; y<CELLS_PER_ROW; y++)
		{
			if (cell_bombs[x][y] == -1)
			{
				continue;
			}
			if (y-1 >= 0)
			{
				for (int k=-1; k<2; k++)
				{
					if (x+k < 0)
					{
						continue;
					}
					else if (x+k >= CELLS_PER_ROW)
					{
						continue;
					}
					if (cell_bombs[x+k][y-1] == -1)
					{
						cell_bombs[x][y]++;
					}
				}
			}
			for (int k=-1; k<2; k++)
			{
				if (x+k < 0)
				{
					continue;
				}
				else if (x+k >= CELLS_PER_ROW)
				{
					continue;
				}
				else if (k==0)
				{
					continue;
				}
				if (cell_bombs[x+k][y] == -1)
				{
					cell_bombs[x][y]++;
				}
			}
			if (y+1 < CELLS_PER_ROW)
			{
				for (int k=-1; k<2; k++)
				{
					if (x+k < 0)
					{
						continue;
					}
					else if (x+k >= CELLS_PER_ROW)
					{
						continue;
					}
					if (cell_bombs[x+k][y+1] == -1)
					{
						cell_bombs[x][y]++;
					}
				}
			}
		}
	}
}
ISR(PCINT2_vect)
{
	cli();
	int state_curr_btn1 = (PINC & (1 << BTN1));
	int state_curr_btn2 = (PINC & (1 << BTN2));
	int state_curr_btn3 = (PINC & (1 << BTN3));
	int state_curr_btn4 = (PINC & (1 << BTN4));
	int state_curr_btn5 = (PINC & (1 << BTN5));
	int state_curr_btn6 = (PINC & (1 << BTN6));
	
	if (lost_game)
	{
		lost_game = 2;
	}
	
	if (BTN1_LAST != state_curr_btn1)
	{
		if (cursor_xpos > 0 && state_curr_btn1 && (!selected_cell) && (!flagged_cell))
		{
			cursor_xpos--;
			_delay_ms(100);	
		}
		BTN1_LAST = state_curr_btn1;
		sei();
	}
	else if (BTN2_LAST != state_curr_btn2)
	{
		if (cursor_xpos < CELLS_PER_ROW-1 && state_curr_btn2 && (!selected_cell) && (!flagged_cell))
		{
			cursor_xpos++;
			_delay_ms(100);
		}
		BTN2_LAST = state_curr_btn2;
		sei();
	}
	else if (BTN3_LAST != state_curr_btn3)
	{
		if (cursor_ypos > 0 && state_curr_btn3)
		{
			cursor_ypos--;
			_delay_ms(100);
		}
		BTN3_LAST = state_curr_btn3;
		sei();
	}
	else if (BTN4_LAST != state_curr_btn4)
	{
		if (cursor_ypos < CELLS_PER_ROW-1 && state_curr_btn4)
		{
			cursor_ypos++;
			_delay_ms(100);
		}
		BTN4_LAST = state_curr_btn4;
		sei();
	}
	else if (BTN5_LAST != state_curr_btn5)
	{
		if (state_curr_btn5)
		{
			selected_cell = true;
			_delay_ms(100);
		}
		BTN5_LAST = state_curr_btn5;
		sei();
	}
	else if (BTN6_LAST != state_curr_btn6)
	{
		if (state_curr_btn6)
		{
			flagged_cell = true;
			_delay_ms(100);
		}
		BTN6_LAST = state_curr_btn6;
		sei();
	}
}
void ResetGame()
{
	fillScreen(BLACK, SLAVEID);

	cursor_xpos = 0;
	cursor_ypos = 0;
	last_cursor_xpos = 0;
	last_cursor_ypos = 0;
	selected_cell = 0;
	flagged_cell = 0;
	lost_game = 0;
	left_bombs = INITIAL_BOMBS_COUNT + (level-1) * LEVEL_INCREASE;

	for (int i=0; i<CELLS_PER_ROW; i++)
	{
		for (int j=0; j<CELLS_PER_ROW; j++)
		{
			cell_bombs[i][j] = 0;
			cell_states[i][j] = 0;
		}
	}
	cell_states[cursor_xpos][cursor_ypos] |= CELL_SELECTED;

	GenerateTable();
	DrawTable();
	UpdateScore();
}
void OpenAllNeighbours(int x, int y)
{
	if (cell_states[x][y] & CELL_FLAGGED)
	{
		return;
	}
	cell_states[x][y] |= CELL_OPENED;
	DrawCell(x, y);
	if (y-1 >= 0)
	{
		for (int k=-1; k<2; k++)
		{
			if (x+k < 0)
			{
				continue;
			}
			else if (x+k >= CELLS_PER_ROW)
			{
				continue;
			}
			if ((cell_bombs[x+k][y-1] == 0) && !(cell_states[x+k][y-1] & CELL_OPENED))
			{
				OpenAllNeighbours(x+k, y-1);
			}
		}
	}
	for (int k=-1; k<2; k++)
	{
		if (x+k < 0)
		{
			continue;
		}
		else if (x+k >= CELLS_PER_ROW)
		{
			continue;
		}
		else if (k==0)
		{
			continue;
		}
		if ((cell_bombs[x+k][y] == 0) && !(cell_states[x+k][y] & CELL_OPENED))
		{
			OpenAllNeighbours(x+k, y);
		}
	}
	if (y+1 < CELLS_PER_ROW)
	{
		for (int k=-1; k<2; k++)
		{
			if (x+k < 0)
			{
				continue;
			}
			else if (x+k >= CELLS_PER_ROW)
			{
				continue;
			}
			if ((cell_bombs[x+k][y+1] == 0) && !(cell_states[x+k][y+1] & CELL_OPENED))
			{
				OpenAllNeighbours(x+k, y+1);
			}
		}
	}
}
int GameWon()
{
	char *text1 = "You Win";
	int offsetx = 10;
	int cnt = 0;

	for (int i=0; i<CELLS_PER_ROW; i++)
	{
		for (int j=0; j<CELLS_PER_ROW; j++)
		{		
			if (cell_states[i][j] & CELL_FLAGGED)
			{
				/* bad guess */
				if (cell_bombs[i][j] != -1)
				{
					return 0;
				}
			}
		}
	}
	fillScreen(BLACK, SLAVEID);
	
	while(*(text1+cnt))
	{
		drawChar(offsetx, 40, *(text1+cnt), RED, BLACK, 2, SLAVEID);
		offsetx += 16;
		cnt++;
	}
	_delay_ms(3000);
	fillScreen(BLACK, SLAVEID);

	return 1;
}
int main()
{
	sei();
	initR();
	setRotation(4, SLAVEID);
	fillScreen(BLACK, SLAVEID);

	_delay_ms(200);
	DDRD = 0xFF;
	PCICR |= (1 << PCIE2);
	PCMSK2 |= (1 << PCINT16) | (1 << PCINT17) | (1 << PCINT18) | (1 << PCINT19) | (1 << PCINT20) | (1 << PCINT21);

	DDRC &= ~(1 << BTN1);
	DDRC &= ~(1 << BTN2);
	DDRC &= ~(1 << BTN3);
	DDRC &= ~(1 << BTN4);
	DDRC &= ~(1 << BTN5);
	DDRC &= ~(1 << BTN6);

	PORTC |= (1 << BTN1);
	PORTC |= (1 << BTN2);
	PORTC |= (1 << BTN3);
	PORTC |= (1 << BTN4);
	PORTC |= (1 << BTN5);
	PORTC |= (1 << BTN6);

	_delay_ms(200);

	ResetGame();	
	while(1)
	{
		_delay_ms(10);
		/* button pressed after game over, reset the whole deal */
		if (lost_game == 2)
		{
			level=1;
			ResetGame();		
			continue;
		}
		else if (lost_game == 1)
		{
			continue;
		}
		if ((last_cursor_xpos != cursor_xpos) || (last_cursor_ypos != cursor_ypos))
		{
			cell_states[last_cursor_xpos][last_cursor_ypos] &= ~(CELL_SELECTED);
			DrawCell(last_cursor_xpos, last_cursor_ypos);

			last_cursor_xpos = cursor_xpos;
			last_cursor_ypos = cursor_ypos;

			cell_states[cursor_xpos][cursor_ypos] |= CELL_SELECTED;
			DrawCell(cursor_xpos, cursor_ypos);
		}
		if (selected_cell)
		{
			selected_cell=false;
			cell_states[cursor_xpos][cursor_ypos] &= ~(CELL_SELECTED);
			/* Remove flag if opening flagged cell */
			if (cell_states[cursor_xpos][cursor_ypos] & CELL_FLAGGED)
			{
				cell_states[cursor_xpos][cursor_ypos] &= ~CELL_FLAGGED;
				left_bombs++;
			}
			cell_states[cursor_xpos][cursor_ypos] |= CELL_OPENED;

			DrawCell(cursor_xpos, cursor_ypos);
			if (cell_bombs[cursor_xpos][cursor_ypos] == -1)
			{
				cell_states[cursor_xpos][cursor_ypos] |= CELL_BOMB_ACTIVATED;
				DrawCell(cursor_xpos, cursor_ypos);
				for (int i=0; i<CELLS_PER_ROW; i++)
				{
					for (int j=0; j<CELLS_PER_ROW; j++)
					{
						if (i == cursor_xpos && j == cursor_ypos)
						{
							continue;
						}

						/* check flagged bombs */
						if (cell_states[i][j] & CELL_FLAGGED)
						{
							/* bad flag, remove it */
							if (cell_bombs[i][j]!=-1)
							{
								cell_states[i][j] &= ~CELL_FLAGGED;
							}
							cell_states[i][j] |= CELL_OPENED;
						}
						else if (cell_bombs[i][j]==-1)
						{
							cell_states[i][j] |= CELL_BOMB_FLAGGED;
							DrawCell(i, j);
						}
						else
						{
							cell_states[i][j] |= CELL_OPENED;
						}
						DrawCell(i, j);
					}
				}
				lost_game = 1;
				cli();//no more intreupts here
				_delay_ms(1000);
				sei();
				continue;
			}
			else if (cell_bombs[cursor_xpos][cursor_ypos] == 0)
			{
				OpenAllNeighbours(cursor_xpos, cursor_ypos);
			}
		}
		else if (flagged_cell)
		{
			flagged_cell=false;
			cell_states[cursor_xpos][cursor_ypos] &= ~(CELL_SELECTED);

			if (cell_states[cursor_xpos][cursor_ypos] & CELL_FLAGGED)
			{
				left_bombs++;
				cell_states[cursor_xpos][cursor_ypos] &= ~CELL_FLAGGED;
				DrawCell(cursor_xpos, cursor_ypos);
				UpdateScore();
			}
			else if (left_bombs)
			{
				left_bombs--;
				cell_states[cursor_xpos][cursor_ypos] |= CELL_FLAGGED;
				DrawCell(cursor_xpos, cursor_ypos);
				UpdateScore();
			}
			if (left_bombs == 0)
			{
				if (GameWon())
				{
					level++;
					ResetGame();
				}
			}
		}
	}
}